@@ -17,11 +17,23 @@ module Sequel
     include Metaprogramming
 
     # Array of supported database adapters
-    ADAPTERS = %w'ado db2 dbi do firebird informix jdbc mysql odbc openbase oracle postgres sqlite'.collect{|x| x.to_sym}
+    ADAPTERS = %w'ado amalgalite db2 dbi do firebird informix jdbc mysql odbc openbase oracle postgres sqlite'.collect{|x| x.to_sym}
 
     SQL_BEGIN = 'BEGIN'.freeze
     SQL_COMMIT = 'COMMIT'.freeze
+    SQL_RELEASE_SAVEPOINT = 'RELEASE SAVEPOINT autopoint_%d'.freeze
     SQL_ROLLBACK = 'ROLLBACK'.freeze
+    SQL_ROLLBACK_TO_SAVEPOINT = 'ROLLBACK TO SAVEPOINT autopoint_%d'.freeze
+    SQL_SAVEPOINT = 'SAVEPOINT autopoint_%d'.freeze
+    
+    TRANSACTION_BEGIN = 'Transaction.begin'.freeze
+    TRANSACTION_COMMIT = 'Transaction.commit'.freeze
+    TRANSACTION_ROLLBACK = 'Transaction.rollback'.freeze
+    
+    POSTGRES_DEFAULT_RE = /\A(?:B?('.*')::[^']+|\((-?\d+(?:\.\d+)?)\))\z/
+    MSSQL_DEFAULT_RE = /\A(?:\(N?('.*')\)|\(\((-?\d+(?:\.\d+)?)\)\))\z/
+    MYSQL_TIMESTAMP_RE = /\ACURRENT_(?:DATE|TIMESTAMP)?\z/
+    STRING_DEFAULT_RE = /\A'(.*)'\z/
 
     # The identifier input method to use by default
     @@identifier_input_method = nil
@@ -69,8 +81,8 @@ module Sequel
     def initialize(opts = {}, &block)
       @opts ||= opts
       
-      @single_threaded = opts.include?(:single_threaded) ? opts[:single_threaded] : @@single_threaded
-      @schemas = nil
+      @single_threaded = opts.include?(:single_threaded) ? typecast_value_boolean(opts[:single_threaded]) : @@single_threaded
+      @schemas = {}
       @default_schema = opts.include?(:default_schema) ? opts[:default_schema] : default_schema_default
       @prepared_statements = {}
       @transactions = []
@@ -98,7 +110,7 @@ module Sequel
         begin
           Sequel.require "adapters/#{scheme}"
         rescue LoadError => e
-          raise AdapterNotFound, "Could not load #{scheme} adapter:\n  #{e.message}"
+          raise Sequel.convert_exception_class(e, AdapterNotFound)
         end
         
         # make sure we actually loaded the adapter
@@ -116,7 +128,8 @@ module Sequel
     
     # Connects to a database.  See Sequel.connect.
     def self.connect(conn_string, opts = {}, &block)
-      if conn_string.is_a?(String)
+      case conn_string
+      when String
         if match = /\A(jdbc|do):/o.match(conn_string)
           c = adapter_class(match[1].to_sym)
           opts = {:uri=>conn_string}.merge(opts)
@@ -125,13 +138,16 @@ module Sequel
           scheme = uri.scheme
           scheme = :dbi if scheme =~ /\Adbi-/
           c = adapter_class(scheme)
-          uri_options = {}
+          uri_options = c.send(:uri_to_options, uri)
           uri.query.split('&').collect{|s| s.split('=')}.each{|k,v| uri_options[k.to_sym] = v} unless uri.query.to_s.strip.empty?
-          opts = c.send(:uri_to_options, uri).merge(uri_options).merge(opts)
+          uri_options.entries.each{|k,v| uri_options[k] = URI.unescape(v) if v.is_a?(String)}
+          opts = uri_options.merge(opts)
         end
-      else
+      when Hash
         opts = conn_string.merge(opts)
         c = adapter_class(opts[:adapter] || opts['adapter'])
+      else
+        raise Error, "Sequel::Database.connect takes either a Hash or a String, given: #{conn_string.inspect}"
       end
       # process opts a bit
       opts = opts.inject({}) do |m, kv| k, v = *kv
@@ -219,9 +235,10 @@ module Sequel
     
     ### Instance Methods ###
 
-    # Executes the supplied SQL statement string.
+    # Runs the supplied SQL statement string on the database server.
+    # Alias for run.
     def <<(sql)
-      execute_ddl(sql)
+      run(sql)
     end
     
     # Returns a dataset from the database. If the first argument is a string,
@@ -414,6 +431,14 @@ module Sequel
       @quote_identifiers = @opts.include?(:quote_identifiers) ? @opts[:quote_identifiers] : (@@quote_identifiers.nil? ? quote_identifiers_default : @@quote_identifiers)
     end
     
+    # Runs the supplied SQL statement string on the database server. Returns nil.
+    # Options:
+    # * :server - The server to run the SQL on.
+    def run(sql, opts={})
+      execute_ddl(sql, opts)
+      nil
+    end
+    
     # Returns a new dataset with the select method invoked.
     def select(*args, &block)
       dataset.select(*args, &block)
@@ -438,16 +463,12 @@ module Sequel
       quoted_name = quote_schema_table(table)
       opts = opts.merge(:schema=>sch) if sch && !opts.include?(:schema)
 
-      @schemas.delete(quoted_name) if opts[:reload] && @schemas
-      return @schemas[quoted_name] if @schemas && @schemas[quoted_name]
-
-      @schemas ||= Hash.new do |h,k|
-        quote_name = quote_schema_table(k)
-        h[quote_name] if h.include?(quote_name)
-      end
+      @schemas.delete(quoted_name) if opts[:reload]
+      return @schemas[quoted_name] if @schemas[quoted_name]
 
       cols = schema_parse_table(table_name, opts)
       raise(Error, 'schema parsing returned no columns, table probably doesn\'t exist') if cols.nil? || cols.empty?
+      cols.each{|_,c| c[:ruby_default] = column_schema_to_ruby_default(c[:default], c[:type])}
       @schemas[quoted_name] = cols
     end
 
@@ -460,20 +481,21 @@ module Sequel
     def synchronize(server=nil, &block)
       @pool.hold(server || :default, &block)
     end
+    
+    # Whether the database and adapter support savepoints, false by default
+    def supports_savepoints?
+      false
+    end
 
     # Returns true if a table with the given name exists.  This requires a query
     # to the database unless this database object already has the schema for
     # the given table name.
     def table_exists?(name)
-      if @schemas && @schemas[name]
+      begin 
+        from(name).first
         true
-      else
-        begin 
-          from(name).first
-          true
-        rescue
-          false
-        end
+      rescue
+        false
       end
     end
     
@@ -484,29 +506,21 @@ module Sequel
       true
     end
 
-    # A simple implementation of SQL transactions. Nested transactions are not 
-    # supported - calling #transaction within a transaction will reuse the 
-    # current transaction. Should be overridden for databases that support nested 
-    # transactions.
-    def transaction(opts={})
+    # Starts a database transaction.  When a database transaction is used,
+    # either all statements are successful or none of the statements are
+    # successful.  Note that MySQL MyISAM tabels do not support transactions.
+    #
+    # The following options are respected:
+    #
+    # * :server  - The server to use for the transaction
+    # * :savepoint - Whether to create a new savepoint for this transaction,
+    #   only respected if the database adapter supports savepoints.  By
+    #   default Sequel will reuse an existing transaction, so if you want to
+    #   use a savepoint you must use this option.
+    def transaction(opts={}, &block)
       synchronize(opts[:server]) do |conn|
-        return yield(conn) if @transactions.include?(Thread.current)
-        log_info(begin_transaction_sql)
-        conn.execute(begin_transaction_sql)
-        begin
-          @transactions << Thread.current
-          yield(conn)
-        rescue Exception => e
-          log_info(rollback_transaction_sql)
-          conn.execute(rollback_transaction_sql)
-          transaction_error(e)
-        ensure
-          unless e
-            log_info(commit_transaction_sql)
-            conn.execute(commit_transaction_sql)
-          end
-          @transactions.delete(Thread.current)
-        end
+        return yield(conn) if already_in_transaction?(conn, opts)
+        _transaction(conn, &block)
       end
     end
     
@@ -520,10 +534,8 @@ module Sequel
       meth = "typecast_value_#{column_type}"
       begin
         respond_to?(meth, true) ? send(meth, value) : value
-      rescue ArgumentError, TypeError => exp
-        e = Sequel::InvalidValue.new("#{exp.class} #{exp.message}")
-        e.set_backtrace(exp.backtrace)
-        raise e
+      rescue ArgumentError, TypeError => e
+        raise Sequel.convert_exception_class(e, InvalidValue)
       end
     end
     
@@ -554,6 +566,66 @@ module Sequel
     
     private
     
+    # Internal generic transaction method.  Any exception raised by the given
+    # block will cause the transaction to be rolled back.  If the exception is
+    # not Sequel::Rollback, the error will be reraised. If no exception occurs
+    # inside the block, the transaction is commited.
+    def _transaction(conn)
+      begin
+        add_transaction
+        t = begin_transaction(conn)
+        yield(conn)
+      rescue Exception => e
+        rollback_transaction(t) if t
+        transaction_error(e)
+      ensure
+        begin
+          commit_transaction(t) unless e
+        rescue Exception => e
+          raise_error(e, :classes=>database_error_classes)
+        ensure
+          remove_transaction(t)
+        end
+      end
+    end
+    
+    # Add the current thread to the list of active transactions
+    def add_transaction
+      th = Thread.current
+      if supports_savepoints?
+        unless @transactions.include?(th)
+          th[:sequel_transaction_depth] = 0
+          @transactions << th
+        end
+      else
+        @transactions << th
+      end
+    end    
+
+    # Whether the current thread/connection is already inside a transaction
+    def already_in_transaction?(conn, opts)
+      @transactions.include?(Thread.current) && (!supports_savepoints? || !opts[:savepoint])
+    end
+    
+    # SQL to start a new savepoint
+    def begin_savepoint_sql(depth)
+      SQL_SAVEPOINT % depth
+    end
+
+    # Start a new database transaction on the given connection.
+    def begin_transaction(conn)
+      if supports_savepoints?
+        th = Thread.current
+        depth = th[:sequel_transaction_depth]
+        conn = transaction_statement_object(conn) if respond_to?(:transaction_statement_object, true)
+        log_connection_execute(conn, depth > 0 ? begin_savepoint_sql(depth) : begin_transaction_sql)
+        th[:sequel_transaction_depth] += 1
+      else
+        log_connection_execute(conn, begin_transaction_sql)
+      end
+      conn
+    end
+    
     # SQL to BEGIN a transaction.
     def begin_transaction_sql
       SQL_BEGIN
@@ -576,17 +648,93 @@ module Sequel
         obj.respond_to?(:empty?) ? obj.empty? : false
       end
     end
+    
+    # Convert the given default, which should be a database specific string, into
+    # a ruby object.
+    def column_schema_to_ruby_default(default, type)
+      return if default.nil?
+      orig_default = default
+      if database_type == :postgres and m = POSTGRES_DEFAULT_RE.match(default)
+        default = m[1] || m[2]
+      end
+      if database_type == :mssql and m = MSSQL_DEFAULT_RE.match(default)
+        default = m[1] || m[2]
+      end
+      if [:string, :blob, :date, :datetime, :time, :enum].include?(type)
+        if database_type == :mysql
+          return if [:date, :datetime, :time].include?(type) && MYSQL_TIMESTAMP_RE.match(default)
+          orig_default = default = "'#{default.gsub("'", "''").gsub('\\', '\\\\')}'"
+        end
+        return unless m = STRING_DEFAULT_RE.match(default)
+        default = m[1].gsub("''", "'")
+      end
+      res = begin
+        case type
+        when :boolean
+          case default 
+          when /[f0]/i
+            false
+          when /[t1]/i
+            true
+          end
+        when :string, :enum
+          default
+        when :blob
+          Sequel::SQL::Blob.new(default)
+        when :integer
+          Integer(default)
+        when :float
+          Float(default)
+        when :date
+          Sequel.string_to_date(default)
+        when :datetime
+          DateTime.parse(default)
+        when :time
+          Sequel.string_to_time(default)
+        when :decimal
+          BigDecimal.new(default)
+        end
+      rescue
+        nil
+      end
+    end
+   
+    # SQL to commit a savepoint
+    def commit_savepoint_sql(depth)
+      SQL_RELEASE_SAVEPOINT % depth
+    end
+
+    # Commit the active transaction on the connection
+    def commit_transaction(conn)
+      if supports_savepoints?
+        depth = Thread.current[:sequel_transaction_depth]
+        log_connection_execute(conn, depth > 1 ? commit_savepoint_sql(depth-1) : commit_transaction_sql)
+      else
+        log_connection_execute(conn, commit_transaction_sql)
+      end
+    end
 
     # SQL to COMMIT a transaction.
     def commit_transaction_sql
       SQL_COMMIT
     end
+    
+    # Method called on the connection object to execute SQL on the database,
+    # used by the transaction code.
+    def connection_execute_method
+      :execute
+    end
 
     # The default options for the connection pool.
     def connection_pool_default_options
       {}
     end
     
+    # Which transaction errors to translate, blank by default.
+    def database_error_classes
+      []
+    end
+    
     # The default value for default_schema.
     def default_schema_default
       nil
@@ -614,6 +762,13 @@ module Sequel
     def input_identifier_meth
       dataset.method(:input_identifier)
     end
+    
+    # Log the given SQL and then execute it on the connection, used by
+    # the transaction code.
+    def log_connection_execute(conn, sql)
+      log_info(sql)
+      conn.send(connection_execute_method, sql)
+    end
 
     # Return a dataset that uses the default identifier input and output methods
     # for this database.  Used when parsing metadata so that column symbols are
@@ -648,9 +803,7 @@ module Sequel
     # and traceback.
     def raise_error(exception, opts={})
       if !opts[:classes] || Array(opts[:classes]).any?{|c| exception.is_a?(c)}
-        e = DatabaseError.new("#{exception.class} #{exception.message}")
-        e.set_backtrace(exception.backtrace)
-        raise e
+        raise Sequel.convert_exception_class(exception, opts[:disconnect] ? DatabaseDisconnectError : DatabaseError)
       else
         raise exception
       end
@@ -660,12 +813,33 @@ module Sequel
     def remove_cached_schema(table)
       @schemas.delete(quote_schema_table(table)) if @schemas
     end
+    
+    # Remove the current thread from the list of active transactions
+    def remove_transaction(conn)
+      th = Thread.current
+      @transactions.delete(th) if !supports_savepoints? || ((th[:sequel_transaction_depth] -= 1) <= 0)
+    end
 
     # Remove the cached schema_utility_dataset, because the identifier
     # quoting has changed.
     def reset_schema_utility_dataset
       @schema_utility_dataset = nil
     end
+    
+    # SQL to rollback to a savepoint
+    def rollback_savepoint_sql(depth)
+      SQL_ROLLBACK_TO_SAVEPOINT % depth
+    end
+
+    # Rollback the active transaction on the connection
+    def rollback_transaction(conn)
+      if supports_savepoints?
+        depth = Thread.current[:sequel_transaction_depth]
+        log_connection_execute(conn, depth > 1 ? rollback_savepoint_sql(depth-1) : rollback_transaction_sql)
+      else
+        log_connection_execute(conn, rollback_transaction_sql)
+      end
+    end
 
     # Split the schema information from the table
     def schema_and_table(table_name)
@@ -682,28 +856,28 @@ module Sequel
     # integer, string, date, datetime, boolean, and float.
     def schema_column_type(db_type)
       case db_type
-      when /\Atinyint/io
-        Sequel.convert_tinyint_to_bool ? :boolean : :integer
       when /\Ainterval\z/io
         :interval
-      when /\A(character( varying)?|varchar|text)/io
+      when /\A(character( varying)?|n?(var)?char|n?text)/io
         :string
-      when /\A(int(eger)?|bigint|smallint)/io
+      when /\A(int(eger)?|(big|small|tiny)int)/io
         :integer
       when /\Adate\z/io
         :date
-      when /\A(datetime|timestamp( with(out)? time zone)?)\z/io
+      when /\A((small)?datetime|timestamp( with(out)? time zone)?)\z/io
         :datetime
       when /\Atime( with(out)? time zone)?\z/io
         :time
-      when /\Aboolean\z/io
+      when /\A(boolean|bit)\z/io
         :boolean
       when /\A(real|float|double( precision)?)\z/io
         :float
-      when /\A(((numeric|decimal)(\(\d+,\d+\))?)|money)\z/io
+      when /\A(((numeric|decimal)(\(\d+,\d+\))?)|(small)?money)\z/io
         :decimal
-      when /bytea|blob/io
+      when /bytea|blob|image|(var)?binary/io
         :blob
+      when /\Aenum/
+        :enum
       end
     end
 
@@ -733,8 +907,8 @@ module Sequel
     end
 
     # Raise a database error unless the exception is an Rollback.
-    def transaction_error(e, *classes)
-      raise_error(e, :classes=>classes) unless Rollback === e
+    def transaction_error(e)
+      raise_error(e, :classes=>database_error_classes) unless e.is_a?(Rollback)
     end
 
     # Typecast the value to an SQL::Blob
@@ -761,6 +935,8 @@ module Sequel
         Date.new(value.year, value.month, value.day)
       when String
         Sequel.string_to_date(value)
+      when Hash
+        Date.new(*[:year, :month, :day].map{|x| (value[x] || value[x.to_s]).to_i})
       else
         raise InvalidValue, "invalid value for Date: #{value.inspect}"
       end
@@ -768,14 +944,12 @@ module Sequel
 
     # Typecast the value to a DateTime or Time depending on Sequel.datetime_class
     def typecast_value_datetime(value)
-      raise(Sequel::InvalidValue, "invalid value for Datetime: #{value.inspect}") unless [DateTime, Date, Time, String].any?{|c| value.is_a?(c)}
-      if Sequel.datetime_class === value
-        # Already the correct class, no need to convert
-       value
+      raise(Sequel::InvalidValue, "invalid value for Datetime: #{value.inspect}") unless [DateTime, Date, Time, String, Hash].any?{|c| value.is_a?(c)}
+      klass = Sequel.datetime_class
+      if value.is_a?(Hash)
+        klass.send(klass == Time ? :mktime : :new, *[:year, :month, :day, :hour, :minute, :second].map{|x| (value[x] || value[x.to_s]).to_i})
       else
-        # First convert it to standard ISO 8601 time, then
-        # parse that string using the time class.
-        Sequel.string_to_datetime(Time === value ? value.iso8601 : value.to_s)
+        Sequel.typecast_to_application_timestamp(value)
       end
     end
 
@@ -813,6 +987,9 @@ module Sequel
         value
       when String
         Sequel.string_to_time(value)
+      when Hash
+        t = Time.now
+        Time.mktime(t.year, t.month, t.day, *[:hour, :minute, :second].map{|x| (value[x] || value[x.to_s]).to_i})
       else
         raise Sequel::InvalidValue, "invalid value for Time: #{value.inspect}"
       end
